// This mod file test the bgp.write function by characterizing the Balanced Growth Path of the Solow model. This is done 10000 times in a Monte Carlo loop
// randomizing the initial guess of the nonlinear solver.

var Efficiency
    EfficiencyGrowth
    Population
    PopulationGrowth
    Output
    PhysicalCapitalStock ;

varexo e_x
       e_n ;

parameters alpha
	       delta
	       s
           rho_x
           rho_n
           EfficiencyGrowth_ss
           PopulationGrowth_ss ;

alpha = .33;
delta = .02;
s     = .20;
rho_x = .90;
rho_n = .95;
EfficiencyGrowth_ss = 1.01;
PopulationGrowth_ss = 1.01;

model(use_dll);
    Efficiency = EfficiencyGrowth*Efficiency(-1);
    EfficiencyGrowth/EfficiencyGrowth_ss = (EfficiencyGrowth(-1)/EfficiencyGrowth_ss)^(rho_x)*exp(e_x);
    Population = PopulationGrowth*Population(-1);
    PopulationGrowth/PopulationGrowth_ss = (PopulationGrowth(-1)/PopulationGrowth_ss)^(rho_n)*exp(e_n);
    Output = PhysicalCapitalStock(-1)^alpha*(Efficiency*Population)^(1-alpha);
    PhysicalCapitalStock = (1-delta)*PhysicalCapitalStock(-1) + s*Output;
end;

verbatim;

  if ~isoctave
  % The levenberg-marquardt algorithm is not available in octave's
  % implementation of fsolve, so we skip this check for Octave

    HardCodedLevels = {'Output', 10, 'Population', 1};

    bgp.write(M_, HardCodedLevels{:});

    MC = 100;
    KY = NaN(MC,1);
    GY = NaN(MC,1);
    GK = NaN(MC,1);
    EG = NaN(MC,1);
    EF = NaN(MC,1);

    if matlab_ver_less_than('9.0')
        % See https://fr.mathworks.com/help/optim/ug/current-and-legacy-option-name-tables.html
        options = optimoptions('fsolve','Display','off','Algorithm','levenberg-marquardt','MaxFunEvals',1000000,'MaxIter',100000,'Jacobian','on','TolFun',1e-8,'TolX',1e-8);
    else
        options = optimoptions('fsolve','Display','off','Algorithm','trust-region','MaxFunctionEvaluations',1000000,'MaxIterations',100000,'SpecifyObjectiveGradient',true,'FunctionTolerance',1e-10,'StepTolerance',1e-8);
    end

    reverseStr = '';

    for i=1:MC
      y = 10+(rand(6,1)-.5)*.2;
      y(2) = 1+.02*rand;
      y(4) = 1+.02*rand;
      G = ones(6,1)+.05*randn(6,1);
      y = [y;G];
      [y, id] = bgp.hdl.out(M_, HardCodedLevels, y);
      [y, fval, exitflag] = fsolve(@solow.bgpfunhardcon, y, options);
      y = bgp.hdl.in(M_, HardCodedLevels, y, id);
      if exitflag>0
        KY(i) = y(6)/y(5);
        GY(i) = y(11);
        GK(i) = y(12);
        EG(i) = y(2);
        EF(i) = y(4);
      end

      % Display the progress
      percentDone = 100 * i / MC;
      msg = sprintf('Percent done: %3.1f', percentDone);
      fprintf([reverseStr, msg]);
      reverseStr = repmat(sprintf('\b'), 1, length(msg));
    end

    skipline();
    % Compute the physical capital stock over output ratio along the BGP as
    % a function of the deep parameters...
    theoretical_long_run_ratio = s*EfficiencyGrowth_ss*PopulationGrowth_ss/(EfficiencyGrowth_ss*PopulationGrowth_ss-1+delta);
    % ... and compare to the average ratio in the Monte Carlo
    mu = mean(KY(~isnan(KY)));
    s2 = var(KY(~isnan(KY)));
    nn = sum(isnan(KY))/MC;
    disp(sprintf('Theoretical K/Y BGP ratio = %s.', num2str(theoretical_long_run_ratio)));
    disp(sprintf('mean(K/Y) = %s.', num2str(mu)));
    disp(sprintf('var(K/Y) = %s.', num2str(s2)));
    disp(sprintf('Number of failures: %u (over %u problems).', sum(isnan(KY)), MC));
    if (abs(mu-theoretical_long_run_ratio)>1e-6), error('Average long run K/Y ratio is wrong.'), end
    if (s2>1e-12), warning('Average long run K/Y ratio is wrong.'), end
    dprintf('Proportion of wrong values for GY is %s.', num2str(sum(abs(GY(~isnan(GY))-EfficiencyGrowth_ss*PopulationGrowth_ss)>1e-6)/sum(~isnan(GY))))
    dprintf('Proportion of wrong values for GK is %s.', num2str(sum(abs(GK(~isnan(GK))-EfficiencyGrowth_ss*PopulationGrowth_ss)>1e-6)/sum(~isnan(GK))))
    dprintf('Proportion of wrong values for EG is %s.', num2str(sum(abs(EG(~isnan(EG))-EfficiencyGrowth_ss)>1e-6)/sum(~isnan(EG))))
    dprintf('Proportion of wrong values for EF is %s.', num2str(sum(abs(EF(~isnan(EF))-PopulationGrowth_ss)>1e-6)/sum(~isnan(EF))))
  end

end;
